// WAVEMIX 2007/1/8 ver.

#include <windows.h>

#include "wavemix.h"

typedef struct{
	PWAVEDATA data;

	BOOL playing;
	DWORD pos;
} PLAYDATA, *PPLAYDATA;

#define WAVEMIXER_MAXCHANNELS 16
#define WAVEMIXER_BUFFERSIZEMS 50	// in Millisecond!
#define WAVEMIXER_BUFFERSIZE (22050 * WAVEMIXER_BUFFERSIZEMS / 1000)
#define WAVEMIXER_CALLBACKCLASSNAME "WaveMixCallbackWindow"

static PLAYDATA wm_playdata[16];
static WAVEHDR whdr[2];
static DWORD playingbuffer;
static HWAVEOUT hWaveOut = NULL;

static WAVEFORMATEX wfe;
static struct
{
	DWORD srate;
	WORD bits;
	WORD ch;
}woConfig = {22050, 8, 1};

static HWND hCallbackWnd = NULL;

void wmDebugPrint(HDC hdc)
{
	char tmp[80];
	int i;
	DWORD pb = 0;

	for(i = 0; i < WAVEMIXER_MAXCHANNELS; i++)
	{
		if(wm_playdata[i].playing)
			wsprintf(tmp, "%02d: %07d/%07d", i, wm_playdata[i].pos, wm_playdata[i].data->length);
		else
			wsprintf(tmp, "%02d: --not playing--", i);
		TextOut(hdc, 0, i * 20, tmp, strlen(tmp));
	}

	wsprintf(tmp, "%d", pb);
	TextOut(hdc, 200, 0, tmp, 1);

	{
		HPEN old = SelectObject(hdc, GetStockObject(WHITE_PEN));
		MoveToEx(hdc, 200, 0x80, NULL);
		for(i = 0; i < 256; i++)
		{
			LineTo(hdc, 200 + i, 0x100 - *(BYTE*)(whdr[pb].lpData + (i * WAVEMIXER_BUFFERSIZE/256 )));
		}
		SelectObject(hdc, old);
	}
	pb = (pb + 1) % 2;
	wsprintf(tmp, "%d", pb);
	TextOut(hdc, 500, 0, tmp, 1);
	{
		HPEN old = SelectObject(hdc, GetStockObject(WHITE_PEN));
		MoveToEx(hdc, 500, 0x80, NULL);
		for(i = 0; i < 256; i++)
		{
			LineTo(hdc, 500 + i, 0x100 - *(BYTE*)(whdr[pb].lpData + (i * WAVEMIXER_BUFFERSIZE/256 )));
		}
		SelectObject(hdc, old);
	}

	return;
}

void wmDoneProc(WAVEHDR* wh)
{
	DWORD i;

	if( wh->dwUser == 1 )	// waveOutReset...
		return;

	for(i = 0; i < WAVEMIXER_BUFFERSIZE; i++)
		whdr[playingbuffer].lpData[i] = (BYTE)0x80;

	for(i = 0; i < WAVEMIXER_MAXCHANNELS; i++)
		if(wm_playdata[i].playing)
		{
			DWORD trans = smaller(wm_playdata[i].data->length - wm_playdata[i].pos, WAVEMIXER_BUFFERSIZE);
			DWORD j;

			for(j = 0; j < trans; j++)
			{
				int data = (int)((BYTE)whdr[playingbuffer].lpData[j]) - 0x80;
				data += (signed int)(*(wm_playdata[i].data->data + wm_playdata[i].pos + j)) - 0x80;
				if(data < -0x80)
					data = -0x80;
				if(data > +0x7F)
					data = +0x7F;
				whdr[playingbuffer].lpData[j] = (BYTE)(data + 0x80);
			}

			wm_playdata[i].pos += trans;
			if(wm_playdata[i].pos >= wm_playdata[i].data->length)
				wm_playdata[i].playing = FALSE;
		}

	waveOutWrite(hWaveOut, &whdr[playingbuffer], sizeof(WAVEHDR));

	playingbuffer = (playingbuffer + 1) % 2;

	return;
}

// No longer used: Can't call system function from WaveOutCallback function...
void CALLBACK waveOutProc(HWAVEOUT hwo, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	switch(uMsg)
	{
	case WOM_DONE:
		wmDoneProc((WAVEHDR*)dwParam1);
		break;
	}

	return;
}

void wmStartWave(void)
{
	int i;

	if(!hWaveOut)
		return;

	for(i = 0; i < 2; i++)
		whdr[i].dwUser = 0;

	playingbuffer = 0;
	waveOutWrite(hWaveOut, &whdr[0], sizeof(WAVEHDR));
	waveOutWrite(hWaveOut, &whdr[1], sizeof(WAVEHDR));

	return;
}

static LRESULT CALLBACK CallbackWindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case MM_WOM_DONE:
		wmDoneProc((WAVEHDR*)lParam);
		break;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}

static BOOL wmCreateCallbackWindow(void)
{
	WNDCLASSEX wc = {0};

	if(hCallbackWnd)
		return FALSE;

	wc.cbSize = sizeof(wc);
	wc.hInstance = GetModuleHandle(NULL);
	wc.lpfnWndProc = CallbackWindowProc;
	wc.lpszClassName = WAVEMIXER_CALLBACKCLASSNAME;
	if(!RegisterClassEx(&wc))
		return TRUE;

	if(!(hCallbackWnd = CreateWindowEx(0, wc.lpszClassName, "WaveMixCallbackWindow!", WS_OVERLAPPEDWINDOW & ~WS_SYSMENU, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, wc.hInstance, NULL)))
		return TRUE;

	//ShowWindow(hCallbackWnd, SW_SHOW);

	return FALSE;
}

static void wmDestroyCallbackWindow(void)
{
	if(!hCallbackWnd)
		return;

	DestroyWindow(hCallbackWnd);
	hCallbackWnd = NULL;

	UnregisterClass(WAVEMIXER_CALLBACKCLASSNAME, GetModuleHandle(NULL));

	return;
}

BOOL wmInitSystem(void)
{
	int i;

	// Oh, already opened.
	if(hWaveOut)
		return FALSE;

	wfe.cbSize = sizeof(wfe);
	wfe.wFormatTag = WAVE_FORMAT_PCM;
	wfe.nChannels = woConfig.ch;
	wfe.wBitsPerSample = woConfig.bits;
	wfe.nBlockAlign = wfe.nChannels * wfe.wBitsPerSample/8;
	wfe.nSamplesPerSec = woConfig.srate;
	wfe.nAvgBytesPerSec = wfe.nSamplesPerSec * wfe.nBlockAlign;

	if(wmCreateCallbackWindow())
		return TRUE;

	if(waveOutOpen(&hWaveOut, WAVE_MAPPER, &wfe, (DWORD)hCallbackWnd, 0, CALLBACK_WINDOW))
		return TRUE;

	for(i = 0; i < 2; i++)
	{
		ZeroMemory(&whdr[i], sizeof(WAVEHDR));

		if(!(whdr[i].lpData = malloc(WAVEMIXER_BUFFERSIZE)))
		{
			MessageBox(NULL, "malloc failed", "WaveMixer: Error", MB_ICONERROR | MB_OK);
			return TRUE;
		}

		whdr[i].dwBufferLength = WAVEMIXER_BUFFERSIZE;
		whdr[i].dwFlags = WHDR_BEGINLOOP | WHDR_ENDLOOP;
		whdr[i].dwLoops = 1;
		whdr[i].dwUser = 0;

		// test
		/*
		{
			int j;
			for(j = 0; j < WAVEMIXER_BUFFERSIZE; j++)
			{
				whdr[i].lpData[j] = (j%64/32)*0x40 - 0x20 + 0x80;
			}
		}
		*/
		memset(whdr[i].lpData, 0x80, WAVEMIXER_BUFFERSIZE);

		if(waveOutPrepareHeader(hWaveOut, &whdr[i], sizeof(WAVEHDR)))
		{
			MessageBox(NULL, "waveOutPrepareHeader failed", "WaveMixer: Error", MB_ICONERROR | MB_OK);
			return TRUE;
		}
	}

	for(i = 0; i < WAVEMIXER_MAXCHANNELS; i++)
	{
		wm_playdata[i].playing = FALSE;
	}

	wmStartWave();

	return FALSE;	// NoError
}

LRESULT CALLBACK MemoryIOProc(LPMMIOINFO lpMMIOInfo, UINT uMsg, LPARAM lParam1, LPARAM lParam2)
{
	static BYTE* buffer = NULL;

	switch(uMsg)
	{
	case MMIOM_OPEN:
		//if(buffer)
		//	return -1;

		buffer = (BYTE*)lParam1;	// Wow
		lpMMIOInfo->lDiskOffset = 0;

		return MMSYSERR_NOERROR;
	case MMIOM_READ:
		memcpy((BYTE*)lParam1, buffer + lpMMIOInfo->lDiskOffset, lParam2);
		lpMMIOInfo->lDiskOffset += lParam2;

		return lParam2;
	case MMIOM_CLOSE:
		buffer = 0;

		return MMSYSERR_NOERROR;
	case MMIOM_SEEK:
		MessageBox(NULL, "MMIOM_SEEK NOT IMPLEMENTED", "IOProc", MB_OK);
		return lpMMIOInfo->lDiskOffset;
		/*
           switch (lParam2) {
           case SEEK_SET:
               lpMMIOInfo->lDiskOffset = lParam1;
               break;
   
           case SEEK_CUR:
               lpMMIOInfo->lDiskOffset += lParam1;
               break;
   
           case SEEK_END:
               lpMMIOInfo->lDiskOffset = fileSize - 1 - lParam1;
               break;
           }
           return lpMMIOInfo->lDiskOffset;
		*/
	case MMIOM_WRITE:
		MessageBox(NULL, "MMIOM_WRITE NOT IMPLEMENTED", "IOProc", MB_OK);
		return -1;
	}

	return -1;	// Error
}

PWAVEDATA wmMemoryWave(void* buf)
{
	PWAVEDATA wd;
	MMIOINFO mi = {0};
	MMCKINFO mcp, mc;
	WAVEFORMATEX wfex;
	BYTE* wav;

	if(!(wd = malloc(sizeof(WAVEDATA))))
		return NULL;

	mi.dwFlags = MMIO_READ | MMIO_ALLOCBUF;	// need?
	mi.fccIOProc = (ULONG)NULL;
	mi.pIOProc = (MMIOPROC*)MemoryIOProc;

	if(!(mi.hmmio = mmioOpen((LPSTR)buf, &mi, MMIO_READ)))
	{
		free(wd);
		return NULL;
	}

	mcp.fccType = mmioFOURCC('W', 'A', 'V', 'E');
	if(mmioDescend(mi.hmmio, &mcp, NULL, MMIO_FINDRIFF))
	{
		free(wd);
		mmioClose(mi.hmmio, 0);
		return NULL;
	}

	// Read PCMWAVEFORMAT
	mc.ckid = mmioFOURCC('f', 'm', 't', ' ');
	if(mmioDescend(mi.hmmio, &mc, &mcp, MMIO_FINDCHUNK))
	{
		free(wd);
		mmioClose(mi.hmmio, 0);
		return NULL;
	}
	if(!mc.cksize)	// What!? Chunksize is ZERO!!
	{
		free(wd);
		mmioClose(mi.hmmio, 0);
		return NULL;
	}
	if(mmioRead(mi.hmmio, (char*)&wfex, sizeof(PCMWAVEFORMAT)) == -1)
	{
		free(wd);
		mmioClose(mi.hmmio, 0);
		return NULL;
	}
	if(mmioAscend(mi.hmmio, &mc, 0))
	{
		free(wd);
		mmioClose(mi.hmmio, 0);
		return NULL;
	}

	// Read DATA!
	mc.ckid = mmioFOURCC('d', 'a', 't', 'a');
	if(mmioDescend(mi.hmmio, &mc, &mcp, MMIO_FINDCHUNK))
	{
		free(wd);
		mmioClose(mi.hmmio, 0);
		return NULL;
	}
	if(!mc.cksize)	// What!? Chunksize is ZERO!!
	{
		free(wd);
		mmioClose(mi.hmmio, 0);
		return NULL;
	}
	if(!(wav = malloc(mc.cksize)))
	{
		free(wd);
		mmioClose(mi.hmmio, 0);
		return NULL;
	}
	if(mmioRead(mi.hmmio, wav, mc.cksize) == -1)
	{
		free(wav);
		free(wd);
		mmioClose(mi.hmmio, 0);
		return NULL;
	}
	wd->length = mc.cksize;
	if(mmioAscend(mi.hmmio, &mc, 0))
	{
		free(wav);
		free(wd);
		mmioClose(mi.hmmio, 0);
		return NULL;
	}

	mmioClose(mi.hmmio, 0);

	if(	wfex.nSamplesPerSec == woConfig.srate &&
		wfex.wBitsPerSample == woConfig.bits &&
		wfex.nChannels == woConfig.ch)
	{	// Oh, it just meat! MEAT! FRESHMEAT!!
		wd->data = wav;
		return wd;
	}

	/*
	// Convert data
	if(!(wd->data = malloc(wd->length * wfex.nSamplesPerSec/woConfig.srate * wfex.wBitsPerSample/woConfig.bits * wfex.nChannels/woConfig.ch)))
	{
		free(wav);
		free(wd);
		return NULL;
	}

	if(wfex.nChannels != woConfig.ch)
	{
	}
	*/
	wd->data = wav;	// oops nothing converted

	return wd;
}

PWAVEDATA wmLoadWave(LPCSTR fname)
{
	PWAVEDATA wd;
	BYTE* wave;
	HANDLE hwf;
	DWORD read = 0;
	DWORD size;

	if(!(hwf = CreateFile(fname, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)))
		return NULL;

	size = GetFileSize(hwf, NULL);

	if(!(wave = malloc(size)))
	{
		CloseHandle(hwf);
		return NULL;
	}

	if(!ReadFile(hwf, wave, size, &read, NULL))
	{
		CloseHandle(hwf);
		return NULL;
	}
	if(read != size)
	{
		CloseHandle(hwf);
		return NULL;
	}

	CloseHandle(hwf);

	if(!(wd = wmMemoryWave(wave)))
	{
		free(wave);
		return NULL;
	}

	free(wave);

	return wd;
}

void wmPlayWave(PWAVEDATA wd)
{
	int i;

	for(i = 0; i < WAVEMIXER_MAXCHANNELS; i++)
		if(!wm_playdata[i].playing)
		{
			MMTIME mmt;
			DWORD cur;
			DWORD pb = playingbuffer;
			BOOL over = FALSE;

			wm_playdata[i].data = wd;
			wm_playdata[i].pos = 0;
			wm_playdata[i].playing = TRUE;

			mmt.wType = TIME_BYTES;
			waveOutGetPosition(hWaveOut, &mmt, sizeof(MMTIME));
			if(mmt.wType != TIME_BYTES)	// oops...
				break;
			cur = mmt.u.cb % WAVEMIXER_BUFFERSIZE;
			cur += 1024;
			if(cur >= WAVEMIXER_BUFFERSIZE)
			{
				cur -= WAVEMIXER_BUFFERSIZE;
				pb = (pb + 1) % 2;
				over = TRUE;
			}

			{
				DWORD trans = smaller(wm_playdata[i].data->length - wm_playdata[i].pos, WAVEMIXER_BUFFERSIZE - cur);
				DWORD j;

				for(j = 0; j < trans; j++)
				{
					int data = (int)((BYTE)whdr[pb].lpData[j + cur]) - 0x80;
					data += (signed int)(*(wm_playdata[i].data->data + wm_playdata[i].pos + j)) - 0x80;
					if(data < -0x80)
						data = -0x80;
					if(data > +0x7F)
						data = +0x7F;
					whdr[pb].lpData[j + cur] = (BYTE)(data + 0x80);
				}

				wm_playdata[i].pos += trans;
				if(wm_playdata[i].pos >= wm_playdata[i].data->length)
				{
					wm_playdata[i].playing = FALSE;
					break;
				}
			}
			if(!over)
			{
				DWORD trans = smaller(wm_playdata[i].data->length - wm_playdata[i].pos, WAVEMIXER_BUFFERSIZE);
				DWORD j;

				pb = (pb + 1) % 2;

				for(j = 0; j < trans; j++)
				{
					int data = (int)((BYTE)whdr[pb].lpData[j]) - 0x80;
					data += (signed int)(*(wm_playdata[i].data->data + wm_playdata[i].pos + j)) - 0x80;
					if(data < -0x80)
						data = -0x80;
					if(data > +0x7F)
						data = +0x7F;
					whdr[pb].lpData[j] = (BYTE)(data + 0x80);
				}

				wm_playdata[i].pos += trans;
				if(wm_playdata[i].pos >= wm_playdata[i].data->length)
				{
					wm_playdata[i].playing = FALSE;
					break;
				}
			}

			break;
		}

	return;
}

void wmKillAllWave(void)
{
	int i, j;
	
	for(i = 0; i < 2; i++)
		for(j = 0; j < WAVEMIXER_BUFFERSIZE; j++)
			whdr[i].lpData[j] = (BYTE)0x80;

	for(i = 0; i < WAVEMIXER_MAXCHANNELS; i++)
		wm_playdata[i].playing = FALSE;

	return;
}

void wmStopWave(void)
{
	int i;

	if(!hWaveOut)
		return;

	for(i = 0; i < 2; i++)
		whdr[i].dwUser = 1;

	waveOutReset(hWaveOut);

	return;
}

void wmPauseWaveOut(void)
{
	if(!hWaveOut)
		return;

	waveOutPause(hWaveOut);

	return;
}

void wmUnpauseWaveOut(void)
{
	if(!hWaveOut)
		return;

	waveOutRestart(hWaveOut);

	return;
}

void wmUnloadWave(PWAVEDATA wd)
{
	free(wd->data);
	free(wd);

	return;
}

void wmUninitSystem(void)
{
	int i;

	// oh, already freed.
	if(!hWaveOut)
		return;

	wmStopWave();

	for(i = 0; i < 2; i++)
	{
		waveOutUnprepareHeader(hWaveOut, &whdr[i], sizeof(WAVEHDR));
		free(whdr[i].lpData);
	}

	waveOutClose(hWaveOut);

	hWaveOut = NULL;

	wmDestroyCallbackWindow();

	return;
}
// WaveMixer END
